#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import requests
import subprocess
import time
import yaml


from novaclient.v1_1 import client


logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class NovaClientException(Exception):
    pass


class CloudInitUserData(object):

    """Maps configuration data into cloud-init user-data.

    This is a container for the data that will ultimately be encoded into a
    cloud-config-archive user-data file.
    """

    def __init__(self):
        super(CloudInitUserData, self).__init__()
        # The objects we will populate before creating a yaml encoding as a
        # cloud-config-archive file
        conf = {}
        # Common pre-requisites:
        conf['apt_update'] = True
        # Default value but we may want to parametrize in the future
        conf['password'] = 'ubuntu'
        # We don't want to be queried to change the password
        conf['chpasswd'] = dict(expire=False)
        # The pre-requisites packages to be able to use adt-run (it's probably
        # a bug in autopkgtest that we need to specify pbuilder here).
        conf['packages'] = ['autopkgtest', 'pbuilder',
                            # We ned rng-tools to generate enough entropy so
                            # gpg can generate its keys.
                            'rng-tools',
                            ]
        self.cloud_config = conf

    def dump(self):
        parts = [{'content': '#cloud-config\n'
                  + yaml.safe_dump(self.cloud_config)}]
        # Wrap the lot into a cloud config archive
        return '#cloud-config-archive\n' + yaml.safe_dump(parts)


def _get_auth_config():
    DEFAULT_CONFIG_YAML = os.path.join(
        os.path.dirname(__file__), '../../unit_config')
    with open(DEFAULT_CONFIG_YAML) as f:
        config = yaml.safe_load(f)
    return config


class TestBed(object):

    def __init__(self, name, flavor, image, status_cb, auth_conf=None):
        if auth_conf is None:
            auth_conf = _get_auth_config()
        self.auth_conf = auth_conf
        self.name = name
        self.flavor = flavor
        self.image = image
        self.status_cb = status_cb
        self.instance = None
        args = [auth_conf.get('auth_user'),
                auth_conf.get('auth_password'),
                auth_conf.get('auth_tenant_name'),
                auth_conf.get('auth_url')]
        kwargs = {'region_name': auth_conf.get('auth_region'),
                  'service_type': 'compute'}
        nova = client.Client(*args, **kwargs)
        self.nova = nova
        self.nova_ssh_key_name = 'test-runner-ssh-key'
        self.test_bed_key_path = None
        self.test_bed_ssh_key_content = None

    def ensure_ssh_key_is_available(self):
        # From the test runner, we need an ssh key that can be used to connect
        # to the test bed. For testing purposes, we rely on self.auth_conf to
        # provide this key.
        if not os.path.exists(self.test_bed_key_path):
            # First time the test runner instance needs to create a test bed
            # instance, we need to create the ssh key pair.
            subprocess.call(
                ['ssh-keygen', '-t', 'rsa',
                 '-f', self.test_bed_key_path, '-N', ''])

        # MISSINGTEST: the key (or the .pub) doesn't exist (related to errors
        # occuring in the ssh-keygen call above but can happen for other
        # reasons) -- vila 2014-01-30
        with open(self.test_bed_key_path + '.pub') as f:
            self.test_bed_ssh_key_content = f.read()

        # Register the key in nova
        if not self.nova.keypairs.findall(name=self.nova_ssh_key_name):
            self.nova.keypairs.create(
                name=self.nova_ssh_key_name,
                public_key=self.test_bed_ssh_key_content)

    def create_user_data(self):
        self.test_bed_key_path = self.auth_conf.get('auth_testbed_sshkey',
                                                    '~/.ssh/id_rsa')
        self.test_bed_key_path = os.path.expanduser(self.test_bed_key_path)
        self.ensure_ssh_key_is_available()
        ud = CloudInitUserData()
        # MISSINGTEST: Since we tell nova to use that key (see setup()), we may
        # not need to use cloud-init to ensure it's authorized on the test bed
        # anymore. -- vila 2014-01-30
        ud.cloud_config['ssh_authorized_keys'] = [
            self.test_bed_ssh_key_content]
        self.user_data = ud.dump()

    def setup(self):
        flavor = self.nova.flavors.find(name=self.flavor)
        image = self.nova.images.find(name=self.image)
        self.create_user_data()
        self.instance = self.nova.servers.create(
            name=self.name,
            flavor=flavor,image=image,
            userdata=self.user_data,
            keypair=self.nova_ssh_key_name
            )
        self.wait_for_ip()

    def wait_for_ip(self):
        status = self.instance.status
        # FIXME: time_to_get_ip should be a config option -- vila 2014-01-30
        time_to_get_ip = 300  # in seconds so 5 minutes
        timeout_limit = time.time() + time_to_get_ip
        while time.time() < timeout_limit and status != 'ACTIVE':
            # Always query nova to get updated data about the instance
            # MISSINGTEST: What if the instance disappear ? -- vila 2014-01-30
            try:
                server = self.nova.servers.get(self.instance.id)
            except requests.ConnectionError:
                log.warn('Received connection error for {},'
                         ' retrying'.format(self.instance.id))
                time.sleep(5)
                continue
            networks = server.networks.values()
            if networks:
                # The network name is arbitrary, can vary for different clouds
                # but there should be only one IP address defined so we just
                # get the first one and avoid the need for a config option for
                # the network name.
                self.ip = networks[0][0]
                log.info('Got IP {} for {}'.format(self.ip, self.instance.id))
                return
            else:
                log.info('IP not yet available for {}'.format(self.instance.id))
            time.sleep(5)
        # MISSINGTEST: What if the instance doesn't come up after max_tries ?
        if status == 'BUILD':
            raise NovaClientException('Instance never came up')
        else:
            raise NovaClientException('Instance never provided an IP')
        # FIXME: We need to poll 'nova console-log' until we get cloud-init
        # completion before attempting to connect to the testbed.

    def teardown(self):
        self.nova.servers.delete(self.instance.id)

    def ssh(self, command, *args):
        user ='ubuntu'
        host = self.ip
        cmd = ['ssh',
               # FIXME: It would be better to ssh-keygen -f
               # "~/.ssh/known_hosts" -R self.ip once we're done with the test
               # bed to avoid polluting ssh commands stdout, but that will do
               # for now (that's what juju is doing after all ;)
               # -- vila 2014-01-29
               '-oStrictHostKeyChecking=no',
               '-i', self.test_bed_key_path,
               '{}@{}'.format(user, host)]
        if command is not None:
            cmd += [command]
        if args:
            cmd += args
        proc = subprocess.Popen(cmd,
#                                stdout=subprocess.PIPE,
#                                stderr=subprocess.PIPE,
                                )
        out, err = proc.communicate()
        return proc, out, err

    def run_test(self, package):
        # FIXME: We'd better upload a predefined gpg env
        self.ssh('sudo', 'rngd', '-r', '/dev/urandom',
                 '-o', '/dev/random',  '-t',  '1')

        # MISSINGTEST: Unable to locate package (wrong name or whatever)
        self.ssh('sudo', 'apt-get', 'install', package)
        self.ssh('apt-get', 'source', package)
        self.ssh('mkdir', 'results')
        _, out, err = self.ssh(
            'sudo', 'adt-run', '--source', '{}*.dsc'.format(package),
            '--summary', 'summary.log', '--output-dir', 'results',
            '---', 'adt-virt-null')
        import pdb; pdb.set_trace()


# The following are helpers for local manual testsm they should disappear once
# we have proper integration tests.

# Those helpers rely on a unit_config file at the root of the branch (/!\ NEVER
# to be put under version control) containing the nova credentials (in effect,
# doing manually what juju-deployer does on the juju units). An additional
# 'auth_testbed_sshkey' can be defined there if you don't use ~/.ssh/id_rsa.

def test_print_ip():
    precise_image = ('ubuntu-released/ubuntu-precise-12.04'
                     '-amd64-server-20131205-disk1.img')
    saucy_image = ('ubuntu-released/ubuntu-saucy-13.10'
                   '-amd64-server-20140129-disk1.img')
    test_bed = TestBed('vila1', 'm1.small', saucy_image,
                       lambda x: None)
    test_bed.setup()
    print 'IP: {}'.format(test_bed.ip)
#    test_bed.teardown()
    return test_bed


def test_dump_user_data():
    ud = CloudInitUserData()
    print ud.dump()


if __name__ == '__main__':
    test_dump_user_data()
    test_bed = test_print_ip()
    import pdb; pdb.set_trace()
    test_bed.run_test('obfsproxy')
